/*
   Make the definitions of the Comparator classes
   into anonymous inner classes at each call to sort().

   NOTE: This file most closely mimics lines 35-40 of
         the file sort-countires.rkt.
*/
import java.util.*;

public class Comparator_base_v5
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));


      // Call sort() and pass it a Comparator that is an
      // anonymous instance of an anonymous inner class.
      countries.sort(new Comparator<Country>() {
         public int compare(Country c1, Country c2) {
            if (c1.getArea() < c2.getArea()) return -1;
            if (c1.getArea() > c2.getArea()) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         public int compare(Country c1, Country c2) {
            if (c1.getArea() > c2.getArea()) return -1;
            if (c1.getArea() < c2.getArea()) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         public int compare(Country c1, Country c2) {
            if (c1.getPopulation() < c2.getPopulation()) return -1;
            if (c1.getPopulation() > c2.getPopulation()) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         public int compare(Country c1, Country c2) {
            if (c1.getPopulation() > c2.getPopulation()) return -1;
            if (c1.getPopulation() < c2.getPopulation()) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         public int compare(Country c1, Country c2) {
            return c1.getName().compareTo(c2.getName());
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         public int compare(Country c1, Country c2) {
            return c2.getName().compareTo(c1.getName());
         }});
      countries.forEach(System.out::println); System.out.println();
   }
}


class Country
{
   private String name;
   private double area;
   private double population;

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public double getArea() { return this.area; }

   public double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
