/*
   Move the definitions of the Comparator classes
   into the scope of the ComparatorTester_one_functor_v2
   class as static nested classes.
*/
import java.util.Comparator;
import java.util.Arrays;
import java.util.List;

public class Comparator_one_functor_v2
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      // Build the comparison functors using the one parameter comparison factories.
      countries.sort(new CountryComparatorByArea(new FnLT()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparatorByArea(new FnGT()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparatorByPopulation(new FnLT()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparatorByPopulation(new FnGT()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparatorByName(new FnStringLT()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparatorByName(new FnStringGT()));
      countries.forEach(System.out::println);System.out.println();
   }


   static class CountryComparatorByArea implements Comparator<Country> {
      Fn<Double> fn; // functor that holds <, >, <=, or >=

      CountryComparatorByArea(Fn<Double> fn) { this.fn = fn; }

      public int compare(Country c1, Country c2) {
         if ( fn.fn(c1.getArea(), c2.getArea()) ) return -1;
         if (!fn.fn(c1.getArea(), c2.getArea()) ) return  1;
         return 0;
      }
   }

   static class CountryComparatorByPopulation implements Comparator<Country> {
      Fn<Double> fn; // functor that holds <, >, <=, or >=

      CountryComparatorByPopulation(Fn<Double> fn) { this.fn = fn; }

      public int compare(Country c1, Country c2) {
         if ( fn.fn(c1.getPopulation(), c2.getPopulation()) ) return -1;
         if (!fn.fn(c1.getPopulation(), c2.getPopulation()) ) return  1;
         return 0;
      }
   }

   static class CountryComparatorByName implements Comparator<Country> {
      Fn<String> fn;

      CountryComparatorByName(Fn<String> fn) { this.fn = fn; }

      public int compare(Country c1, Country c2) {
         if ( fn.fn(c1.getName(), c2.getName()) ) return -1;
         if (!fn.fn(c1.getName(), c2.getName()) ) return  1;
         return 0;
      }
   }


   // Since we are not using lambdas, and we cannot pass operators
   // as parameters in Java, we need the following interfaces as
   // wrappers for simple functions.
   static abstract class Fn<T> { // create a "functor" type
      public abstract boolean fn(T x, T y);
   }

   static class FnLT extends Fn<Double> {
      public boolean fn(Double x, Double y) {
         return x < y;
      }
   }

   static class FnGT extends Fn<Double> {
      public boolean fn(Double x, Double y) {
         return x > y;
      }
   }

   static class FnStringLT extends Fn<String> {
      @Override
      public boolean fn(String x, String y) {
         return x.compareTo(y) <= -1;
      }
   }

   static class FnStringGT extends Fn<String> {
      @Override
      public boolean fn(String x, String y) {
         return x.compareTo(y) >= +1;
      }
   }
}

class Country
{
   private String name;
   private double area;
   private double population;

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public double getArea() { return this.area; }

   public double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
