/*
   Make the definitions of the Comparator classes
   into anonymous inner classes at each call
   to sort().
*/
import java.util.Comparator;
import java.util.Arrays;
import java.util.List;

public class Comparator_one_functor_v5
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      // Since we are not using lambdas, and we cannot pass operators
      // as parameters in Java, we need the following interface as a
      // wrapper for simple functions.
      abstract class Fn<T> { // create a "functor" type
         public abstract boolean fn(T x, T y);
      }

      // Call sort() and pass it a Comparator that is an
      // anonymous instance of an anonymous inner class.
      // Notice that each comparator also holds a named
      // instance of another anonymous inner class.
      countries.sort(new Comparator<Country>() {
         Fn<Double> fn = new Fn<Double>() {
            public boolean fn(Double x, Double y) {
               return x < y;
            }};
         public int compare(Country c1, Country c2) {
            if ( fn.fn(c1.getArea(), c2.getArea()) ) return -1;
            if (!fn.fn(c1.getArea(), c2.getArea()) ) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         Fn<Double> fn = new Fn<Double>() {
            public boolean fn(Double x, Double y) {
               return x > y;
            }};
         public int compare(Country c1, Country c2) {
            if ( fn.fn(c1.getArea(), c2.getArea()) ) return -1;
            if (!fn.fn(c1.getArea(), c2.getArea()) ) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         Fn<Double> fn = new Fn<Double>() {
            public boolean fn(Double x, Double y) {
               return x < y;
            }};
         public int compare(Country c1, Country c2) {
            if ( fn.fn(c1.getPopulation(), c2.getPopulation()) ) return -1;
            if (!fn.fn(c1.getPopulation(), c2.getPopulation()) ) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         Fn<Double> fn = new Fn<Double>() {
            public boolean fn(Double x, Double y) {
               return x > y;
            }};
         public int compare(Country c1, Country c2) {
            if ( fn.fn(c1.getPopulation(), c2.getPopulation()) ) return -1;
            if (!fn.fn(c1.getPopulation(), c2.getPopulation()) ) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         Fn<String> fn = new Fn<String>() {
            public boolean fn(String x, String y) {
               return x.compareTo(y) <= -1;
            }};
         public int compare(Country c1, Country c2) {
            if ( fn.fn(c1.getName(), c2.getName()) ) return -1;
            if (!fn.fn(c1.getName(), c2.getName()) ) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();

      countries.sort(new Comparator<Country>() {
         Fn<String> fn = new Fn<String>() {
            public boolean fn(String x, String y) {
               return x.compareTo(y) >= +1;
            }};
         public int compare(Country c1, Country c2) {
            if ( fn.fn(c1.getName(), c2.getName()) ) return -1;
            if (!fn.fn(c1.getName(), c2.getName()) ) return  1;
            return 0;
         }});
      countries.forEach(System.out::println); System.out.println();
   }
}

class Country
{
   private String name;
   private double area;
   private double population;

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public double getArea() { return this.area; }

   public double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
