/*
   Move the definitions of the Comparator classes
   into the scope of the main() method as inner
   nested classes.
*/
import java.util.Comparator;
import java.util.Arrays;
import java.util.List;

public class Comparator_two_functor_v4
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      // Since we are not using lambdas, and we cannot pass operators
      // as parameters in Java, we need the following two interfaces
      // as wrappers for simple functions.
      abstract class Fn<T> { // create a "functor" type
         public abstract boolean fn(T x, T y);
      }

      abstract class Get<T> { // create a "functor" type
         public abstract T get(Country c);
      }

      // The comparator is now an anonymous inner class.
      // NOTE 1: The inner class is anonymous, but the
      // comparator objects are named comp1 through comp6
      // (so we have named instances of an anonymous class).
      // NOTE 2: Each comparator also holds two named instances
      // of two more anonymous inner classes. We are using
      // nested anonymous inner classes.
      Comparator<Country> comp1 = new
         Comparator<Country>() {
            Fn<Double> fn = new Fn<Double>() {
               public boolean fn(Double x, Double y) {
                  return x < y;
               }};
            Get<Double> get = new Get<Double>() {
               public Double get(Country c) {
                  return c.getArea();
               }};
            public int compare(Country c1, Country c2) {
               if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
               if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
               return 0;
            }};

      Comparator<Country> comp2 = new
         Comparator<Country>() {
            Fn<Double> fn = new Fn<Double>() {
               public boolean fn(Double x, Double y) {
                  return x > y;
               }};
            Get<Double> get = new Get<Double>() {
               public Double get(Country c) {
                  return c.getArea();
               }};
            public int compare(Country c1, Country c2) {
               if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
               if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
               return 0;
            }};

      Comparator<Country> comp3 = new
         Comparator<Country>() {
            Fn<Double> fn = new Fn<Double>() {
               public boolean fn(Double x, Double y) {
                  return x < y;
               }};
            Get<Double> get = new Get<Double>() {
               public Double get(Country c) {
                  return c.getPopulation();
               }};
            public int compare(Country c1, Country c2) {
               if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
               if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
               return 0;
            }};

      Comparator<Country> comp4 = new
         Comparator<Country>() {
            Fn<Double> fn = new Fn<Double>() {
               public boolean fn(Double x, Double y) {
                  return x > y;
               }};
            Get<Double> get = new Get<Double>() {
               public Double get(Country c) {
                  return c.getPopulation();
               }};
            public int compare(Country c1, Country c2) {
               if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
               if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
               return 0;
            }};

      Comparator<Country> comp5 = new
         Comparator<Country>() {
            Fn<String> fn = new Fn<String>() {
               public boolean fn(String x, String y) {
                  return x.compareTo(y) <= -1;
               }};
            Get<String> get = new Get<String>() {
               public String get(Country c) {
                  return c.getName();
               }};
            public int compare(Country c1, Country c2) {
               if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
               if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
               return 0;
            }};

      Comparator<Country> comp6 = new
         Comparator<Country>() {
            Fn<String> fn = new Fn<String>() {
               public boolean fn(String x, String y) {
                  return x.compareTo(y) >= +1;
               }};
            Get<String> get = new Get<String>() {
               public String get(Country c) {
                  return c.getName();
               }};
            public int compare(Country c1, Country c2) {
               if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
               if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
               return 0;
            }};

      countries.sort(comp1);
      countries.forEach(System.out::println); System.out.println();

      countries.sort(comp2);
      countries.forEach(System.out::println); System.out.println();

      countries.sort(comp3);
      countries.forEach(System.out::println); System.out.println();

      countries.sort(comp4);
      countries.forEach(System.out::println); System.out.println();

      countries.sort(comp5);
      countries.forEach(System.out::println); System.out.println();

      countries.sort(comp6);
      countries.forEach(System.out::println); System.out.println();
   }
}

class Country
{
   private String name;
   private double area;
   private double population;

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public double getArea() { return this.area; }

   public double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
