/*
   Create single parameter comparator factories
   that are as close to Scheme's as possible.
*/
import java.util.*;
import java.util.function.*;

public class Java8_Comparator_one_functor
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      BiPredicate<Double, Double> lt = (x,y) -> x < y;
      BiPredicate<Double, Double> gt = (x,y) -> x > y;
      BiPredicate<String, String> string_lt = (x,y) -> x.compareTo(y) <= -1;;
      BiPredicate<String, String> string_gt = (x,y) -> x.compareTo(y) >= +1;


      class CmpCountriesByArea implements Comparator<Country> {
         private BiPredicate<Double, Double> cmp;

         CmpCountriesByArea(BiPredicate<Double, Double> cmp) {
            this.cmp = cmp;
         }

         public int compare(Country c1, Country c2) {
            if (cmp.test(c1.getArea(), c2.getArea())) return -1;
            if (cmp.test(c2.getArea(), c1.getArea())) return  1;
            return 0;
         }
      }

      class CmpCountriesByPopulation implements Comparator<Country> {
         private BiPredicate<Double, Double> cmp;

         CmpCountriesByPopulation(BiPredicate<Double, Double> cmp) {
            this.cmp = cmp;
         }

         public int compare(Country c1, Country c2) {
            if (cmp.test(c1.getPopulation(), c2.getPopulation())) return -1;
            if (cmp.test(c2.getPopulation(), c1.getPopulation())) return  1;
            return 0;
         }
      }

      class CmpCountriesByName implements Comparator<Country> {
         private BiPredicate<String, String> cmp;

         CmpCountriesByName(BiPredicate<String, String> cmp) {
            this.cmp = cmp;
         }

         public int compare(Country c1, Country c2) {
            if (cmp.test(c1.getName(), c2.getName())) return -1;
            if (cmp.test(c2.getName(), c1.getName())) return  1;
            return 0;
         }
      }


      countries.sort(new CmpCountriesByArea(lt));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountriesByArea(gt));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountriesByPopulation(lt));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountriesByPopulation(gt));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountriesByName(string_gt));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountriesByName(string_gt));
      countries.forEach(System.out::println);System.out.println();
   }
}


class Country
{
   private String name;
   private Double area;        // note the small change here
   private Double population;  // note the small change here

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public Double getArea() { return this.area; }

   public Double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
