/*
   Create a doubly parameterized comparator factory
   that is as close to Scheme's as possible.
*/
import java.util.*;
import java.util.function.*;

public class Java8_Comparator_two_functor
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      BiPredicate<Double, Double> lt = (x,y) -> x < y;
      BiPredicate<Double, Double> gt = (x,y) -> x > y;
      BiPredicate<String, String> string_lt = (x,y) -> x.compareTo(y) <= -1;;
      BiPredicate<String, String> string_gt = (x,y) -> x.compareTo(y) >= +1;

      Function<Country,Double> selectA = Country::getArea;
      Function<Country,Double> selectP = Country::getPopulation;
      Function<Country,String> selectN = Country::getName;

      class CmpCountries<T> implements Comparator<Country> {
         BiPredicate<T, T>   cmp;  // functor that holds <, >, <=, or >=
         Function<Country,T> get;  // get method

         CmpCountries(BiPredicate<T, T> cmp, Function<Country,T> get) {
            this.cmp = cmp;
            this.get = get;
         }

         public int compare(Country c1, Country c2) {
            if (cmp.test(get.apply(c1), get.apply(c2))) return -1;
            if (cmp.test(get.apply(c2), get.apply(c1))) return  1;
            return 0;
         }
      }

      countries.sort(new CmpCountries<Double>(lt, selectA));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountries<Double>(gt, selectA));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountries<Double>(lt, selectP));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountries<Double>(gt, selectP));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountries<String>(string_gt, selectN));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CmpCountries<String>(string_gt, selectN));
      countries.forEach(System.out::println);System.out.println();
   }
}


class Country
{
   private String name;
   private Double area;        // note the small change here
   private Double population;  // note the small change here

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public Double getArea() { return this.area; }

   public Double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
