/*
   Create a doubly parameterized comparator factory
   that is as close to Scheme's as possible.
*/
import java.util.*;
import java.util.function.*;

public class Java8_Comparator_two_functor_v2
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      BiPredicate<Double, Double> lt = (x,y) -> x < y;
      BiPredicate<Double, Double> gt = (x,y) -> x > y;
      BiPredicate<String, String> string_lt = (x,y) -> x.compareTo(y) <= -1;;
      BiPredicate<String, String> string_gt = (x,y) -> x.compareTo(y) >= +1;

      Function<Country,Double> selectA = Country::getArea;
      Function<Country,Double> selectP = Country::getPopulation;
      Function<Country,String> selectN = Country::getName;

      BiFunction<BiPredicate<Double,Double>,
                 Function<Country,Double>,
                 Comparator<? super Country>>
         makeCmp = (cmp, get) -> (c1, c2) -> {
            if (cmp.test(get.apply(c1), get.apply(c2))) return -1;
            if (cmp.test(get.apply(c2), get.apply(c1))) return  1;
            return 0;
         };

      countries.sort( makeCmp.apply(lt, selectA) );
      countries.forEach(System.out::println);System.out.println();

      countries.sort( makeCmp.apply(gt, selectA) );
      countries.forEach(System.out::println);System.out.println();

      countries.sort( makeCmp.apply(lt, selectP) );
      countries.forEach(System.out::println);System.out.println();

      countries.sort( makeCmp.apply(gt, selectP) );
      countries.forEach(System.out::println);System.out.println();

      BiFunction<BiPredicate<String,String>,
                 Function<Country,String>,
                 Comparator<? super Country>>
         makeCmpS = (cmp, get) -> (c1, c2) -> {
            if (cmp.test(get.apply(c1), get.apply(c2))) return -1;
            if (cmp.test(get.apply(c2), get.apply(c1))) return  1;
            return 0;
         };

      countries.sort( makeCmpS.apply(string_lt, selectN) );
      countries.forEach(System.out::println);System.out.println();

      countries.sort( makeCmpS.apply(string_gt, selectN) );
      countries.forEach(System.out::println);System.out.println();


/*
//  https://stackoverflow.com/questions/22588518/lambda-expression-and-generic-method

      <T> BiFunction<BiPredicate<T,T>,
                     Function<Country,T>,
                     Comparator<? super Country>>
             makeCmp = (cmp, get) -> (c1, c2) -> {
                if (cmp.test(get.apply(c1), get.apply(c2))) return -1;
                if (cmp.test(get.apply(c2), get.apply(c1))) return  1;
                return 0;
             };

      countries.sort( makeCmp(lt, selectA) );
      countries.sort( makeCmp(gt, selectA) );
      countries.sort( makeCmp(lt, selectP) );
      countries.sort( makeCmp(gt, selectP) );
      countries.sort( makeCmp(string_lt, selectN) );
      countries.sort( makeCmp(string_gt, selectN) );
*/
   }
}


class Country
{
   private String name;
   private Double area;        // note the small change here
   private Double population;  // note the small change here

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public Double getArea() { return this.area; }

   public Double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
