#lang racket

(define countries
   '(("Uruguay"  176220   3.4)
     ("Thailand" 513000    67)
     ("Belgium"   30528    11)
     ("Zaire"   2345410  46.5) 
     ("Belize"    22966  0.334)))

; recursive insertion sort
(define sort
  (lambda (lst cmp)  ; cmp is a function parameter
    (letrec   ; similar to begin
        ((insert
          (lambda (x lst)  ; nested function (invariant: the list is sorted)
            (cond
              ((null? lst) (cons x '()))
              ((cmp x (car lst)) (cons x lst)) ; cmp is a non-local reference
              (else                            ; that is lexically scoped
               (cons (car lst) (insert x (cdr lst)))))))
         (nestedSort
          (lambda (lst)                ; another nested function
            (cond ((null? lst) '())
            (else (insert (car lst) (nestedSort (cdr lst))))))))
      (nestedSort lst))))


; Three get (or selector) functions for country objects.
(define getN (lambda (country) (car   country)))  ; get name
(define getA (lambda (country) (cadr  country)))  ; get area
(define getP (lambda (country) (caddr country)))  ; get population


; Build the comparison functions without using factory functions.
(sort countries (lambda (c1 c2) (< (getA c1) (getA c2))))
(sort countries (lambda (c1 c2) (> (getA c1) (getA c2))))
(sort countries (lambda (c1 c2) (< (getP c1) (getP c2))))
(sort countries (lambda (c1 c2) (> (getP c1) (getP c2))))
(sort countries (lambda (c1 c2) (string<? (getN c1) (getN c2))))
(sort countries (lambda (c1 c2) (string>? (getN c1) (getN c2))))


; One paramater comparison factories.
(define makeCmpByArea              ; makeCmpByArea returns a closure
  (lambda (fn)                     ; fn is a function parameter
    (lambda (c1 c2)                ; a function return value
      (fn (getA c1) (getA c2)))))  ; in this lambda, fn is a non-local reference

(define makeCmpByPopulation        ; makeCmpByPopulation returns a closure
  (lambda (fn)                     ; fn is a function parameter
    (lambda (c1 c2)                ; a function return value
      (fn (getP c1) (getP c2)))))  ; in this lambda, fn is a non-local reference

(define makeCmpByName              ; makeCmpByName returns a closure
  (lambda (fn)                     ; fn is a function parameter
    (lambda (c1 c2)                ; a function return value
      (fn (getN c1) (getN c2)))))  ; in this lambda, fn is a non-local reference

; Build the comparison functions using the one parameter comparison factories.
(sort countries (makeCmpByArea <))
(sort countries (makeCmpByArea >))
(sort countries (makeCmpByPopulation <))
(sort countries (makeCmpByPopulation >))
(sort countries (makeCmpByName string<?))
(sort countries (makeCmpByName string>?))


; Two parameter comparison factory.
(define makeCmp                    ; makeCmp returns a closure
  (lambda (fn get)                 ; fn and get are function parameters
    (lambda (c1 c2)                ; makeCmp has a function return value
      (fn (get c1) (get c2)))))    ; in this lambda, fn, get are non-local references

; Build the comparison functions using the two parameter comparison factory.
(sort countries (makeCmp < getA))
(sort countries (makeCmp > getA))
(sort countries (makeCmp < getP))
(sort countries (makeCmp > getP))
(sort countries (makeCmp string<? getN))
(sort countries (makeCmp string>? getN))
